/* PLEASE READ THE 'NOTES' SECTION OF ORACLEAUTH.H BEFORE YOU MODIFY THIS SOURCE
   BETTER YET ASK REGAN BEFORE MODIFYING THIS SOURCE.
*/
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>

#include "oracleauth.h"
#include "oalib.h"
#include "ld_sha.h"

/* Oracle headers... */
#include <oratypes.h>
#include <ociapr.h>
#include <ocidem.h>
#include <ocidfn.h>

typedef struct cmd_list { char *cmd; int params; char *desc; } Cmd, *pCmd;
Cmd cmds[] = {
	{"check",	 2,		"<user> <pass>"},
	{"lookup",	 1,		"<user>"},
	{"set",		-1,		"<user> <pass>|(NULL) [label=\"value\"]"},
	{"del",		 1,		"<user>"},
	{"search",	-1,		"<string> [-from n] [-max m]"},
	{"version",	 0,		""},
	{"help",	 0,		""},
	{"exit",	 0,		""},
	{"quit",	 0,		""},
	{NULL,0}
};

/* This array MUST end with an empty string entry
   enum in oracleauth.h MUST match entries in this table */
ISet config[] = {
	{"loglevel",		"info",		0},
	{"debug",			"false",	0},
	{"domain",			"",			0},
	{"oracle_user",		"",			0},
	{"oracle_pass",		"",			0},
	{"oracle_db",		"users",	0},
	{"oracle_table",	"maildb",	0},
	{"field_user",		"username",	0},
	{"field_pass",		"password",	0},
	{"field_domain",	"",			0},
	{"field_translate",	"",			1},
	{"use_sha",			"true",		0},
	{"append_domain",	"true",		0},
	{"",""}
};

char workpath[BFSZ]="", oabasename[BFSZ]="";
int interactive = TRUE;
pLog xlog = NULL;
pIni ini = NULL;

/* Oracle variables... */
Cda_Def cda, cursor;
ub1 hda[HDA_SIZE];
Lda_Def lda;

int cmd_is(char *string, int *params);
int cmd_call(char *string, int *used);
int pre_args(int argc, char *argv[]);
int post_args(int argc, char *argv[]);
void exename(char *name);
/*
int getinfo(char *user, char *info, int bfsz);
int setinfo(char *user, char *info);
int delinfo(char *user);
*/
void sig_dump(int signo);
void sig_init(void);

typedef struct user_info_struct {
	char field[BFSZ];
	char value[BFSZ];
	sb4 lfield, dbsize, dsize;
	sb2 dbtype, prec, scale, nullok;
	struct user_info_struct *next;
} UInfo, *pUInfo;

int oracle_connect(void);
void oracle_disconnect(void);
char *oracle_error(Cda_Def *cursor);
void oracle_fields_free(pUInfo *fields);
pUInfo oracle_fields(int *count, int *field_user_ok, int *field_pass_ok, int *field_domain_ok);
int oracle_translate(char *from, char *to, int bfsz);
int oracle_istranslated(char *translation, char *of, int bfsz);
int oracle_setpass(char *pass, char *bf, int bfsz);
int oracle_pass_compare(char *xstored, char *entered);

int oracle_check(char *username, char *password);
int oracle_set(char *username, char *password, char *info);
int oracle_del(char *username);
int oracle_search(char *string, int from, int max);

void response(char *format, ...) {
    char temp[BFSZ], *ptr;
    va_list arg;
    
    va_start(arg,format);
    vsnprintf(temp,BFSZ-1,format,arg);
    for (ptr = temp + strlen(temp) - 1; *ptr == '\r' || *ptr == '\n'; ptr--) *ptr = '\0';
    imsg("OUTPUT[%s]",temp);
    printf("%s\n",temp);
    fflush(stdout);
}

int main(int argc, char *argv[]) {
	char bf[BFSZ], logfile[BFSZ], inifile[BFSZ], *ptr;
	int i;

	sig_init();

	/* Load args that have effect before ini and log initialisation */
	if ( !pre_args(argc, argv) ) {
		response("-DEAD Error with cmd line parameters, %s\n", get_error());
		fflush(stdout);
		return TRUE;
	}

	/* Work out my program's name i.e. "unixauth" */
	exename(argv[0]);

	/* Generate path to logfile and open for append */
	ncpy(logfile, workpath, BFSZ);
	attach_path(logfile, oabasename, BFSZ);
	ncat(logfile, ".log", BFSZ);

	xlog = log_open(logfile);
	if ( !xlog ) {
		response("-DEAD Unable to append to logfile '%s', %s\n", logfile, get_error());
		fflush(stdout);
		return TRUE;
	}

	imsg("OracleAuth %s started",VERSION);

	/* Generate path to inifile and load */
	ncpy(inifile, workpath, BFSZ);
	attach_path(inifile, oabasename, BFSZ);
	ncat(inifile, ".ini", BFSZ);

	ini = ini_init(config);
	if ( !ini_load(ini, inifile) ) {
		response("-DEAD Failed to load config settings, %s\n", get_error());
		fflush(stdout);
		return TRUE;
	}

	/* Set loglevel LOG_INFO is default */
	ptr = ini_value(ini, I_LOGLEVEL);
	if ( !strcmp(ptr, "debug") ) {
		dmsg("Setting loglevel to debug");
		log_setlvl(xlog, LOG_DEBUG);
	}
	else if ( !strcmp(ptr, "error") )
		log_setlvl(xlog, LOG_ERROR);

	/* Load remaining args */
	if ( !post_args(argc, argv) ) {
		response("-DEAD Error with cmd line parameters, %s\n", get_error());
		fflush(stdout);
		return TRUE;
	}

	while ( interactive ) {
		if ( !fgets(bf, BFSZ, stdin) ) break;
		if ( !cmd_call(bf, &i) ) {
			response("-ERR %s\n", get_error());
			fflush(stdout);
		}
	}

	/* Clear data, exit */
	oracle_disconnect();
	ini_free(&ini);
	log_free(&xlog);
	return FALSE;
}

int cmd_is(char *string, int *params) {
	int index;

	for ( index = 0; cmds[index].cmd != NULL; index++ ) {
		if ( !strcmp(cmds[index].cmd, string) ) {
			*params = cmds[index].params;
			return TRUE;
		}
	}
	return FALSE;
}
/* WARNING: cmd_call will place \0 characters into string */
int cmd_call(char *string, int *used) {
	char *args[CMD_MAX_ARGS],*p;
	int argc,inquot;
	
	for(p = string+strlen(string); p != string && iseol(*p); p--) *p = '\0';
	imsg("INPUT[%s]",string);

	/* Chop string into command + args */
	inquot = FALSE;
    for ( argc = 0, p = string; argc < CMD_MAX_ARGS && *p; p++) {
		args[argc] = p;
		for(; *p; p++) {
			if (*p == '\"') inquot = !inquot;
			if (inquot) continue;
			if (isspace(*p) || iseol(*p)) break;
		}
		if (!*p) {
			dmsg("cmd param %d = %s", argc, args[argc++]);
			break;
		}
		*p = '\0';
		dmsg("cmd param %d = %s", argc, args[argc++]);
    }

	/* Check we got at least the command */
	if ( argc == 0 ) {
		set_error("No command string given to parse");
		return FALSE;
	}

	/* Take one off cos args[0] is command name */
	argc--;

	dmsg("Command [%s], args=%d", args[0], argc);

	if ( !strcmp(args[0], "check") ) {
		if ( argc < 2 ) {
			if (argc == 1) set_error("%s not found or password incorrect",args[1]);
			else set_error("The [check] command requires two parameters");
			return FALSE;
		}
		*used = 2;
		oracle_check(args[1], args[2]);
		return TRUE;
	}
	else if ( !strcmp(args[0], "lookup") ) {
		if ( argc < 1 ) {
			set_error("The [lookup] command requires a parameter");
			return FALSE;
		}
		*used = 1;
		oracle_check(args[1], NULL);
		return TRUE;
	}
	else if ( !strcmp(args[0], "set") ) {
		char info[LBFSZ];
		int index;

		if ( argc < 2 ) {
			set_error("The [set] command requires at least two parameters");
			return FALSE;
		}

		*used = 1;

		info[0] = '\0';
		for ( index = 3; index <= argc; index++ ) {
			if ( !strchr(args[index], '=') ) break;
			if ( info[0] ) ncat(info, " ", LBFSZ);
			ncat(info, args[index], LBFSZ);
			(*used)++;
		}
		
		oracle_set(args[1], args[2], info);
		return TRUE;
	}
	else if ( !strcmp(args[0], "del") ) {
		if ( argc < 1 ) {
			set_error("The [del] command requires a parameter");
			return FALSE;
		}
		*used = 1;
		oracle_del(args[1]);
		return TRUE;
	}
	else if ( !strcmp(args[0], "search") ) {
		int from = 0, max = -1;

		if ( argc < 1 ) {
			set_error("The [search] command requires a parameter");
			return FALSE;
		}

		*used = 1;
		if ( argc >= 3 ) {
			if ( !strncmp(args[2], "-from", 5) ) {
				from = atoi(args[3]);
				(*used)++;
			}		
			if ( !strncmp(args[2], "-max", 4) ) {
				max = atoi(args[3]);
				(*used)++;
			}
		}
		if ( argc >= 5 ) {
			if ( !strncmp(args[4], "-from", 5) ) {
				from = atoi(args[5]);
				(*used)++;
			}
			if ( !strncmp(args[4], "-max", 4) ) {
				max = atoi(args[5]);
				(*used)++;
			}
		}

		/* Treat 1 and 0 as same thing meaning first user */
		if ( from == 1 ) from = 0;
		
		oracle_search(args[1], from, max);
		return TRUE;
	}
	else if ( !strcmp(args[0], "version") ) {
		response("+OK OracleAuth %s\n", VERSION);
		fflush(stdout);
		*used = 0;
		return TRUE;
	}
	else if ( !strcmp(args[0], "help") ) {
		int index;

		response("+DATA OracleAuth %s help\n", VERSION);
		for ( index = 0; cmds[index].cmd; index++ )
			response("+DATA %s %s\n", cmds[index].cmd, cmds[index].desc);
		response("+OK done\n");
		fflush(stdout);
		*used = 0;
		return TRUE;
	}
	else if ( !strcmp(args[0], "quit") || !strcmp(args[0], "exit") ) {
		interactive = FALSE;
		return TRUE;
	}
	set_error("Command [%s] not found", args[0]);
	return FALSE;
}

int pre_args(int argc, char *argv[]) {
	int i;

	if ( argc == 1 ) return TRUE;
	for ( i = 1; i < argc; i++ ) {
		if ( !strcmp(argv[i], "-path") ) {			
			if ( argc < (i+1) ) {
				set_error("Expected parameter following '%s'", argv[i]);
				return FALSE;
			}
			ncpy(workpath, argv[i+1], BFSZ-1);
		}
	}
	return TRUE;
}

int post_args(int argc, char *argv[]) {
	char cmd_string[BFSZ];
	int i, j, param, used;

	if ( argc == 1 ) return TRUE;
	for ( i = 1; i < argc; i++ ) {
		dmsg("Got arg [%s]", argv[i]);
		if ( !strcmp(argv[i], "-debug") ) {
			log_setlvl(xlog, LOG_DEBUG);
		}
		else {
			if ( cmd_is(&argv[i][1], &param) ) {
				interactive = FALSE;

				ncpy(cmd_string, &argv[i][1], BFSZ);
				for ( j = i+1; j < argc && param != 0; j++, param-- ) {
					ncat(cmd_string, " ", BFSZ);
					ncat(cmd_string, argv[j], BFSZ);
				}
				if ( !cmd_call(cmd_string, &used) ) {
					response("-ERR %s\n", get_error());
					fflush(stdout);
				}
				i += used;
			}
		}
	}
	return TRUE;
}

/* Works out my program name i.e. "ntauth" */
void exename(char *name) {
	char *ptr;
	int i;

	/* Find last sep (if present) */
	for ( ptr = name + strlen(name) - 1; ptr != name; ptr-- ) {
		if ( issep(*ptr) ) {
			ptr++;
			break;
		}
	}

	/* Copy into oabasename */
	for ( i = 0; i < BFSZ && !iseol(*ptr); i++ )
		oabasename[i] = *ptr++;
	oabasename[i] = '\0';

	/* Remove last .ext from end of filename (if present) */
	ptr = oabasename + strlen(oabasename) - 1;
	while ( ptr != oabasename ) {
		if ( *ptr == '.' ) {
			*ptr = '\0';
			break;
		}
		ptr--;
	}
}

/*
int getinfo(char *user, char *info, int bfsz) {
	char file[BFSZ], hash[NA_HASH_SZ];
	FILE *fp;

	ncpy(file, ini_value(ini, I_USER_WORKAREA), BFSZ);
	attach_path(file, na_hash(hash, user), BFSZ);
	attach_path(file, user, BFSZ);
	ncat(file, ".dat", BFSZ);
	fp = fopen(file, "r");
	if ( !fp && errno != ENOENT ) {
		set_error("Failed to open '%s' for read, %s", file, strerror(errno));
		return FALSE;
	}
	info[0] = '\0';
	if ( !fp ) return TRUE;
	fgets(info, BFSZ, fp);
	fclose(fp);
	strip_eol(info);
	return TRUE;
}

int setinfo(char *user, char *info) {
	char file[BFSZ], hash[NA_HASH_SZ];
	FILE *fp;

	ncpy(file, ini_value(ini, I_USER_WORKAREA), BFSZ);
	attach_path(file, na_hash(hash, user), BFSZ);
	if ( !na_mkdir(file) ) return FALSE;
	attach_path(file, user, BFSZ);
	ncat(file, ".dat", BFSZ);
	fp = fopen(file, "w");
	if ( !fp ) {
		set_error("Failed to create '%s', %s", file, strerror(errno));
		return FALSE;
	}
	fprintf(fp, "%s\n", info);
	fclose(fp);
	return TRUE;
}

int delinfo(char *user) {
	char file[BFSZ], hash[NA_HASH_SZ];

	ncpy(file, ini_value(ini, I_USER_WORKAREA), BFSZ);
	attach_path(file, na_hash(hash, user), BFSZ);
	if ( !na_mkdir(file) ) return FALSE;
	attach_path(file, user, BFSZ);
	ncat(file, ".dat", BFSZ);
	remove(file);
	return TRUE;
}
*/

void sig_dump(int signo) {
	signal(SIGTERM, SIG_DFL);
	signal(SIGFPE,  SIG_DFL);
	signal(SIGABRT, SIG_DFL);
	signal(SIGILL,  SIG_DFL);
	signal(SIGSEGV, SIG_DFL);
	emsg("Signal %d caught, stopping...", signo);
	log_flush(xlog);
	log_free(&xlog);
	abort();
}

void sig_init(void) {
	signal(SIGTERM,   sig_dump);
	signal(SIGFPE,    sig_dump);
	signal(SIGABRT,   sig_dump);
	signal(SIGILL,    sig_dump);
	signal(SIGSEGV,   sig_dump);
}

void emsg(char *message, ...) {
	char bf[BFSZ];
	va_list args;
	int t;

	if ( !xlog ) return ;
	va_start(args, message);
	vsnprintf(bf, BFSZ, message, args);
	va_end(args);
	time(&t);
	fprintf(xlog->fp, "%.*s ERROR: %s\n", 24, asctime(localtime(&t)), bf);
}

void imsg(char *message, ...) {
	char bf[BFSZ];
	va_list args;
	int t;

	if ( !xlog || xlog->level < LOG_INFO ) return ;
	va_start(args, message);
	vsnprintf(bf, BFSZ, message, args);
	va_end(args);
	time(&t);
	fprintf(xlog->fp, "%.*s INFO : %s\n", 24, asctime(localtime(&t)), bf);
}

void dmsg(char *message, ...) {
	char bf[BFSZ];
	va_list args;
	int t;

	if ( !xlog || xlog->level < LOG_DEBUG ) return ;
	va_start(args, message);
	vsnprintf(bf, BFSZ, message, args);
	va_end(args);
	time(&t);
	fprintf(xlog->fp, "%.*s DEBUG: %s\n", 24, asctime(localtime(&t)), bf);
}

static int oracle_connected = FALSE;

int oracle_connect(void) {	
	sword status;

	if ( oracle_connected == TRUE ) return TRUE;
	memset(hda, '\0', HDA_SIZE);
	status = olog(&lda, hda, ini_value(ini, I_ORACLE_USER), -1, ini_value(ini, I_ORACLE_PASS), -1, ini_value(ini, I_ORACLE_DB), -1, OCI_LM_DEF);
	if ( status == 0 ) oracle_connected = TRUE;
	return oracle_connected;
}

void oracle_disconnect(void) {
	if ( oracle_connected == FALSE ) return;
	if ( ologof(&lda) ) emsg("Logoff failed, %s", oracle_error(NULL));
}

char *oracle_error(Cda_Def *cursor) {
	static char errmsg[BFSZ];

	if ( cursor != NULL ) oerhms(&lda, cursor->rc, errmsg, (sword)BFSZ);
	else oerhms(&lda, lda.rc, errmsg, (sword)BFSZ);
	return errmsg;
}

void oracle_fields_free(pUInfo *fields) {
	pUInfo ptr, next;

	for ( ptr = *fields; ptr; ptr = next ) {
		next = ptr->next;
		myfree(ptr);
	}
}

pUInfo oracle_fields(int *count, int *field_user_ok, int *field_pass_ok, int *field_domain_ok) {
	static int user_ok = FALSE, pass_ok = FALSE, domain_ok = FALSE, field_count = 0;
	static pUInfo fields = NULL, next, ptr;	
	char sql_statement[BFSZ];	
	int pos;

	if ( fields != NULL ) {
		*field_user_ok = user_ok;
		*field_pass_ok = pass_ok;
		*field_domain_ok = domain_ok;
		*count = field_count;
		return fields;
	}

	if ( oopen(&cda, &lda, NULL, -1, -1, NULL, -1) ) {
		response("-ERR %s", oracle_error(&cda));
		return NULL;
	}

	snprintf(sql_statement, BFSZ, "SELECT * FROM %s", ini_value(ini, I_ORACLE_TABLE));
	dmsg("SQL(%s)",sql_statement);

	if ( oparse(&cda, sql_statement, -1, TRUE, 2) ) {
		response("-ERR %s", oracle_error(&cda));
		return NULL;
    }

	if (!ini_value(ini, I_FIELD_DOMAIN)[0]) {
		dmsg("field_domain not set, domain not stored in a seperate column");
		if ( field_domain_ok ) *field_domain_ok = TRUE;
	}

	for ( pos = 1, ptr = NULL; ; pos++ ) {
		next = calloc(1, sizeof(UInfo));
		next->lfield = BFSZ;
		if ( odescr(&cda, pos, &next->dbsize, &next->dbtype, (text *)&next->field, &next->lfield, &next->dsize, &next->prec, &next->scale, &next->nullok) ) {
			myfree(next);
			if ( cda.rc == 1007 ) break;
			response("-ERR %s", oracle_error(&cda));
			oracle_fields_free(&fields);
			return NULL;
		}
		next->field[next->lfield] = '\0';
		strlwr(next->field);
		if ( !(*field_user_ok) && !stricmp(next->field, ini_value(ini, I_FIELD_USER)) ) {
			*field_user_ok = TRUE;
		}
		if ( !(*field_pass_ok) && !stricmp(next->field, ini_value(ini, I_FIELD_PASS)) ) {
			*field_pass_ok = TRUE;
		}
		if ( !(*field_domain_ok) && !stricmp(next->field, ini_value(ini, I_FIELD_DOMAIN)) ) {
			*field_domain_ok = TRUE;
		}
		if ( !ptr ) ptr = fields = next;
		ptr->next = next;
		ptr = next;
	}

	field_count = *count = pos - 1;

	if ( oclose(&cda) ) {
		response("-ERR %s", oracle_error(&cda));
		oracle_fields_free(&fields);
		return NULL;
	}

	user_ok = *field_user_ok;
	pass_ok = *field_pass_ok;
	domain_ok = *field_domain_ok;

	return fields;
}

/* Looks at "field_translate a b" and turn 'from' if it matches a into b */
int oracle_translate(char *from, char *to, int bfsz) {
	char temp[BFSZ], *setting, *ptr;
	int index = 0;

	setting = ini_value_n(ini, I_FIELD_TRANSLATE, index);
	for ( ; setting[0]; setting = ini_value_n(ini, I_FIELD_TRANSLATE, ++index) ) {
		ncpy(temp, setting, BFSZ);
		for ( ptr = temp; *ptr && *ptr != ' ' && *ptr != '\t'; ptr++ );
		if ( !*ptr ) {
			emsg("field_translate setting [%s] incorrect", temp);
			continue;
		}
		*ptr++ = '\0';
		if ( strcmp(from, temp) != 0 ) continue;
		for ( ; *ptr == ' ' || *ptr == '\t'; ptr++ );
		ncpy(to, ptr, bfsz);
		return TRUE;
	}
	return FALSE;
}

/* Looks at "field_translate a b" and turn 'translation' if it matches b into a */
int oracle_istranslated(char *translation, char *of, int bfsz) {
	char temp[BFSZ], *setting, *ptr;
	int index = 0;

	setting = ini_value_n(ini, I_FIELD_TRANSLATE, index);
	for ( ; setting[0]; setting = ini_value_n(ini, I_FIELD_TRANSLATE, ++index) ) {
		ncpy(temp, setting, BFSZ);
		for ( ptr = temp; *ptr && *ptr != ' ' && *ptr != '\t'; ptr++ );
		if ( !*ptr ) {
			emsg("field_translate setting [%s] incorrect", temp);
			continue;
		}
		*ptr++ = '\0';
		for ( ; *ptr == ' ' || *ptr == '\t'; ptr++ );
		if ( strcmp(translation, ptr) != 0 ) continue;
		ncpy(of, temp, bfsz);
		return TRUE;
	}
	return FALSE;
}

int oracle_setpass(char *pass, char *bf, int bfsz) {
	if (!pass || !bf || bf <= 0) return FALSE;
	if (stricmp(ini_value(ini,I_USE_SHA),"true")==0) {
		strcpy(bf,"{ssha}");
		if (!create_ssha_base64_digest(pass,&bf[6],bfsz-10)) {
			emsg("SSHA encrypt failed, %s",get_sha_error());
			return FALSE;
		}
		return TRUE;
	}
	ncpy(bf,pass,bfsz);
	return TRUE;
}

int oracle_pass_compare(char *xstored, char *entered) {
	char enc[BFSZ],stored[BFSZ];
	int length;

	length = ncpy(stored,xstored,BFSZ);
	if (strncmp(stored,"{ssha}",6) == 0) {
		dmsg("{ssha} encrypted password compare");
		raw_create_ssha(entered,enc,BFSZ-10,stored+6);
		if (strcmp(enc,stored+6) == 0) return TRUE;
	}
	dmsg("Normal password compare");
	return strcmp(stored,entered) == 0;
}

int oracle_getinfo(pUInfo fields, int field_count, char *info, int bfsz) {
	char temp[BFSZ],bf[BFSZ];
	pUInfo ptr;
	int i;

	if (!info) return FALSE; 	
	info[0] = '\0';

	if (!fields) return FALSE;

	/* Generate output info string from user info */
	for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next ) {
		if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_USER)) ) continue;
		if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_PASS)) ) continue;
		if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_DOMAIN)) ) continue;
		/*if ( !strcmp(ptr->value, " ") ) continue;*/
		/* Check if field is translated via setting in ini file */
		if ( oracle_translate(ptr->field, bf, BFSZ) )
			snprintf(temp, BFSZ, "%s=\"%s\"", bf, ptr->value);
		else
			snprintf(temp, BFSZ, "%s=\"%s\"", ptr->field, ptr->value);
		if ( info[0] != '\0' ) ncat(info, " ", bfsz);
		ncat(info, temp, bfsz);
	}
	return TRUE;
}
int oracle_check(char *username, char *password) {
	char uname[BFSZ], domain[BFSZ], info[BFSZ], temp[BFSZ], sql_statement[BFSZ], bf[BFSZ];
	int i, field_count, user_ok, pass_ok, domain_ok;
	pUInfo fields, ptr;

	/* Strip name + domain to name and domain */
	oa_stripname(username, uname, domain, BFSZ);
	if ( ini_value(ini, I_DOMAIN)[0] ) {
		if ( !domain[0] ) {
			domain[0] = '@';
			ncpy(&domain[1], ini_value(ini, I_DOMAIN), BFSZ-1);
		}
	}
	/*else domain[0] = '\0';*/

	/* Ensure oracle connection */
	if ( !oracle_connect() ) {
		response("-ERR %s", oracle_error(NULL));
		return FALSE;
	}

	/* Find out names of all fields in oracle db, check for username and password */
	user_ok = pass_ok = FALSE;
	fields = oracle_fields(&field_count, &user_ok, &pass_ok, &domain_ok);
	if ( !fields ) return FALSE;
	if ( !user_ok ) {
		response("-ERR The config setting 'field_user' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_USER), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	if ( !pass_ok ) {
		response("-ERR The config setting 'field_pass' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_PASS), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	if ( !domain_ok ) {
		response("-ERR The config setting 'field_domain' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_DOMAIN), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	
	/* Open a cursor into db, this is used for all commands to db */
	if ( oopen(&cursor, &lda, NULL, -1, -1, NULL, -1) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Create comma sep list of all fields for sql_statement */
	temp[0] = '\0';
	for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next ) {
		if ( temp[0] != '\0' ) ncat(temp, ",", BFSZ);
		ncat(temp, ptr->field, BFSZ);
	}
	
	/* Create and send sql command, if (password) then check for password match as well as username */
	/*if ( password ) snprintf(sql_statement, BFSZ, "SELECT %s FROM %s WHERE %s=:uname AND %s=:pwd", temp, ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_FIELD_USER), ini_value(ini, I_FIELD_PASS));*/
	/*else*/
	if (ini_value(ini,I_FIELD_DOMAIN)[0]) {
		snprintf(sql_statement, BFSZ, "SELECT %s FROM %s WHERE %s=:uname AND %s=:dom", temp, ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_FIELD_USER), ini_value(ini, I_FIELD_DOMAIN));
	}
	else {
		snprintf(sql_statement, BFSZ, "SELECT %s FROM %s WHERE %s=:uname", temp, ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_FIELD_USER));
	}	
	dmsg("SQL(%s)",sql_statement);
	if ( oparse(&cursor, sql_statement, -1, TRUE, 2) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Set :uname (declared in sql_statement above) to the one we want */
	if (!ini_value(ini,I_FIELD_DOMAIN)[0] && stricmp(ini_value(ini,I_APPEND_DOMAIN),"true")==0) snprintf(bf, BFSZ, "%s%s", uname, domain);
	else snprintf(bf, BFSZ, "%s", uname);
	
	if ( obndra(&cursor, ":uname", -1, bf, -1, SQLT_STR, -1, NULL, NULL, NULL, 0, NULL, NULL, 0, 0) ) {
		response("-ERR %s%s password wrong or not a valid user\n", uname, domain);
		return FALSE;
	}

	if (ini_value(ini,I_FIELD_DOMAIN)[0]) {
		if ( obndra(&cursor, ":dom", -1, &domain[1], -1, SQLT_STR, -1, NULL, NULL, NULL, 0, NULL, NULL, 0, 0) ) {
			response("-ERR %s%s password wrong or not a valid user\n", uname, domain);
			return FALSE;
		}
	}

	/* if (password) set :pwd (declared in sql_statement above) to the one we want
	if ( password ) {
		if ( obndra(&cursor, ":pwd", -1, password, -1, SQLT_STR, -1, NULL, NULL, NULL, 0, NULL, NULL, 0, 0) ) {
			printf("-ERR %s%s password wrong or not a valid user\n", uname, domain);
			return FALSE;
		}
	}
	*/

	/* Execute sql command */
	if ( oexec(&cursor) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Setup buffers to retrieve results to */
	for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next )
		odefin(&cursor, i+1, ptr->value, sizeof(ptr->value), SQLT_STR, -1, (sb2 *) -1, NULL, 0, 0, NULL, NULL);
	
	/* Get first result, we should only get one, so only get one and ignore any others */
	/* 1405 == NULL returned, 1406 == value truncated : ignore them */
	if ( ofetch(&cursor) && cursor.rc != 1405 && cursor.rc != 1406 ) {
		response("-ERR %s%s password wrong or not a valid user\n", uname, domain);
		return FALSE;
	}

	/* Close cursor into db */
	if ( oclose(&cursor) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Compare password */
	if (password) {
		for (i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next) {
			if (stricmp(ptr->field, ini_value(ini, I_FIELD_PASS)) == 0) break;
		}
		if (!ptr) {
			response("-ERR %s%s password wrong or not a valid user\n", uname, domain);
			return FALSE;
		}
		if (!oracle_pass_compare(ptr->value,password)) {
			response("-ERR %s%s password wrong or not a valid user\n", uname, domain);
			return FALSE;
		}
	}

	oracle_getinfo(fields,field_count,info,BFSZ);

	/* Output result */
	response("+OK %s%s config 0 %s\n", uname, domain, info);
	return TRUE;
}

int oracle_set(char *username, char *password, char *info) {
	char uname[BFSZ], domain[BFSZ], sql_statement[BFSZ], temp[BFSZ], bf[BFSZ], *str;
	int user_ok, pass_ok, domain_ok, field_count, i, j, exists, dosetpass;
	pUInfo fields, ptr;
	
	/* Strip name + domain to name and domain */
	oa_stripname(username, uname, domain, BFSZ);
	if ( ini_value(ini, I_DOMAIN)[0] ) {
		if ( !domain[0] ) {
			domain[0] = '@';
			ncpy(&domain[1], ini_value(ini, I_DOMAIN), BFSZ-1);
		}
	}
	/*else domain[0] = '\0';*/

	/* Ensure oracle connection */
	if ( !oracle_connect() ) {
		response("-ERR %s", oracle_error(NULL));
		return FALSE;
	}

	/* Find out names of all fields in oracle db, check for username and password */
	user_ok = pass_ok = FALSE;
	fields = oracle_fields(&field_count, &user_ok, &pass_ok, &domain_ok);
	if ( !fields ) return FALSE;
	if ( !user_ok ) {
		response("-ERR The config setting 'field_user' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_USER), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	if ( !pass_ok ) {
		response("-ERR The config setting 'field_pass' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_PASS), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	if ( !domain_ok ) {
		response("-ERR The config setting 'field_domain' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_DOMAIN), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}

	/* Open a cursor into db, this is used for all commands to db */
	if ( oopen(&cursor, &lda, NULL, -1, -1, NULL, -1) ) {
		response("-ERR (oopen)%s", oracle_error(&cursor));
		return FALSE;
	}

	/* Create and send sql command, this is a lookup command to check if we're adding a new user or modifying an old one */
	if (ini_value(ini,I_FIELD_DOMAIN)[0]) {
		snprintf(sql_statement, BFSZ, "SELECT %s,%s FROM %s WHERE %s=:uname AND %s=:dom", ini_value(ini, I_FIELD_USER), ini_value(ini, I_FIELD_DOMAIN), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_FIELD_USER), ini_value(ini, I_FIELD_DOMAIN));
	}
	else {
		snprintf(sql_statement, BFSZ, "SELECT %s FROM %s WHERE %s=:uname", ini_value(ini, I_FIELD_USER), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_FIELD_USER));
	}
	dmsg("SQL(%s)",sql_statement);
	if ( oparse(&cursor, sql_statement, -1, TRUE, 2) ) {
		response("-ERR (oparse)%s", oracle_error(&cursor));
		return FALSE;
	}

	/* Set :uname (declared in sql_statement above) to the one we want */
	if (!ini_value(ini,I_FIELD_DOMAIN)[0] && stricmp(ini_value(ini,I_APPEND_DOMAIN),"true")==0) snprintf(bf, BFSZ, "%s%s", uname, domain);
	else snprintf(bf, BFSZ, "%s", uname);

	if ( obndra(&cursor, ":uname", -1, bf, -1, SQLT_STR, -1, NULL, NULL, NULL, 0, NULL, NULL, 0, 0) ) {
		response("-ERR %s%s (obndra(:uname))%s\n", uname, domain, oracle_error(&cursor));
		return FALSE;
	}

	if (ini_value(ini,I_FIELD_DOMAIN)[0]) {
		if ( obndra(&cursor, ":dom", -1, &domain[1], -1, SQLT_STR, -1, NULL, NULL, NULL, 0, NULL, NULL, 0, 0) ) {
			response("-ERR %s%s (obndra(:dom))%s\n", uname, domain, oracle_error(&cursor));
			return FALSE;
		}
	}

	/* Execute sql command */
	if ( oexec(&cursor) ) {
		response("-ERR (oexec)%s", oracle_error(&cursor));
		return FALSE;
	}

	/* Find username field and set recive buffer for the data */
	for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next ) {
		if ( stricmp(ptr->field, ini_value(ini, I_FIELD_USER)) ) {
			odefin(&cursor, 1, ptr->value, sizeof(ptr->value), SQLT_STR, -1, (sb2 *) -1, NULL, 0, 0, NULL, NULL);
		}
		if ( stricmp(ptr->field, ini_value(ini, I_FIELD_DOMAIN)) ) {
			odefin(&cursor, 1, ptr->value, sizeof(ptr->value), SQLT_STR, -1, (sb2 *) -1, NULL, 0, 0, NULL, NULL);
		}
	}
	
	/* Find the user, if they exist */
	if ( !ofetch(&cursor) ) exists = TRUE;
	else exists = FALSE;

	/* Close cursor, so we can open a new one - not sure if you can do more than one operation with same cursor?? */
	if ( oclose(&cursor) ) {
		response("-ERR (oclose)%s", oracle_error(&cursor));
		return FALSE;
	}

	/* Open a cursor into db, this is used for all commands to db */
	if ( oopen(&cursor, &lda, NULL, -1, -1, NULL, -1) ) {
		response("-ERR (oopen)%s", oracle_error(&cursor));
		return FALSE;
	}

	/* Build sql_statement based on user existance and whether we are setting a new password */
	if ( exists ) {
		if ( !strcmp(password, "(NULL)") ) dosetpass = FALSE;
		else dosetpass = TRUE;

		snprintf(sql_statement, BFSZ, "UPDATE %s SET ", ini_value(ini, I_ORACLE_TABLE));
		for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next ) {
			if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_DOMAIN)) ) continue;
			if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_USER)) ) continue;
			if ( !dosetpass && !stricmp(ptr->field, ini_value(ini, I_FIELD_PASS)) ) continue;			
			ncat(sql_statement, ptr->field, BFSZ);
			ncat(sql_statement, "=", BFSZ);
			ncat(sql_statement, ":", BFSZ);
			ncat(sql_statement, ptr->field, BFSZ);
			if ( i != field_count-1 ) ncat(sql_statement, ",", BFSZ);
		}
		ncat(sql_statement, " WHERE ", BFSZ);
		ncat(sql_statement, ini_value(ini, I_FIELD_USER), BFSZ);
		ncat(sql_statement, "=:", BFSZ);
		ncat(sql_statement, ini_value(ini, I_FIELD_USER), BFSZ);
		if (ini_value(ini,I_FIELD_DOMAIN)[0]) {
			ncat(sql_statement, " AND ", BFSZ);
			ncat(sql_statement, ini_value(ini, I_FIELD_DOMAIN), BFSZ);
			ncat(sql_statement, "=:", BFSZ);
			ncat(sql_statement, ini_value(ini, I_FIELD_DOMAIN), BFSZ);
		}
	}
	else {
		snprintf(sql_statement, BFSZ, "INSERT INTO %s VALUES(", ini_value(ini, I_ORACLE_TABLE));
		for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next ) {
			ncat(sql_statement, ":", BFSZ);
			ncat(sql_statement, ptr->field, BFSZ);
			if ( i != field_count-1 ) ncat(sql_statement, ",", BFSZ);
			else ncat(sql_statement, ")", BFSZ);
		}
	}

	dmsg("SQL(%s)",sql_statement);

	/* Send sql_statement to db */
	if ( oparse(&cursor, sql_statement, -1, TRUE, 2) ) {
		response("-ERR (oparse)%s", oracle_error(&cursor));
		return FALSE;
	}

	/* Set values buffers for all user fields, copy values in so exec cmd below will read them */
	for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next ) {
		if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_USER)) ) {
			if (!ini_value(ini,I_FIELD_DOMAIN)[0] && stricmp(ini_value(ini,I_APPEND_DOMAIN),"true")==0)
				snprintf(ptr->value, BFSZ, "%s%s", uname, domain);
			else
				snprintf(ptr->value, BFSZ, "%s", uname);
			
		}
		else if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_DOMAIN)) ) {
			snprintf(ptr->value, BFSZ, "%s", &domain[1]);
		}
		else if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_PASS)) ) {
			if ( exists && !dosetpass ) continue;
			oracle_setpass(password,ptr->value,BFSZ);
			/*ncpy(ptr->value, password, BFSZ);*/
		}
		else {
			if ( !oracle_translate(ptr->field, temp, BFSZ) ) ncpy(temp, ptr->field, BFSZ);
			for ( str = strstr(info, temp); str; str = strstr(str, temp) ) {
				str += strlen(temp);
				if ( *str == '=' ) {
					str += 2;
					break;
				}
			}
			if ( !str ) ncpy(ptr->value, "", BFSZ);
			else {
				for ( j = 0; str[j] && str[j] != '"'; j++ );
				ncpy(ptr->value, str, j+1);
			}
		}
		snprintf(bf, BFSZ, ":%s", ptr->field);
		obndrv(&cursor, bf, -1, ptr->value, strlen(ptr->value)+1, STRING_TYPE, -1, 0, 0, -1, -1);
	}
	
	/* Excecute the command */
	if ( oexec(&cursor) ) {
		response("-ERR (oexec)%s", oracle_error(&cursor));
		return FALSE;
	}

	/* Close cursor */
	if ( oclose(&cursor) ) {
		response("-ERR (oclose)%s", oracle_error(&cursor));
		return FALSE;
	}

	/* Commit changes to db */
	if ( ocom(&lda) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Output */
	response("+OK %s%s Added to the database\n", uname, domain);
	return TRUE;
}

int oracle_del(char *username) {
	char uname[BFSZ], domain[BFSZ], sql_statement[BFSZ], bf[BFSZ];
	int user_ok, pass_ok, domain_ok, field_count;
	pUInfo fields;

	/* Strip name + domain to name and domain */
	oa_stripname(username, uname, domain, BFSZ);
	if ( ini_value(ini, I_DOMAIN)[0] ) {
		if ( !domain[0] ) {
			domain[0] = '@';
			ncpy(&domain[1], ini_value(ini, I_DOMAIN), BFSZ-1);
		}
	}
	/*else domain[0] = '\0';*/

	/* Ensure oracle connection */
	if ( !oracle_connect() ) {
		response("-ERR %s", oracle_error(NULL));
		return FALSE;
	}

	/* Find out names of all fields in oracle db, check for username and password */
	user_ok = pass_ok = FALSE;
	fields = oracle_fields(&field_count, &user_ok, &pass_ok, &domain_ok);
	if ( !fields ) return FALSE;
	if ( !user_ok ) {
		response("-ERR The config setting 'field_user' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_USER), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	if ( !pass_ok ) {
		response("-ERR The config setting 'field_pass' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_PASS), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	if ( !domain_ok ) {
		response("-ERR The config setting 'field_domain' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_DOMAIN), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}

	/* Open a cursor into db, this is used for all commands to db */
	if ( oopen(&cursor, &lda, NULL, -1, -1, NULL, -1) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Build sql_statement */
	if (ini_value(ini,I_FIELD_DOMAIN)[0]) {
		snprintf(sql_statement, BFSZ, "DELETE FROM %s WHERE %s=:uname AND %s=:dom", ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_FIELD_USER), ini_value(ini, I_FIELD_DOMAIN));
	}
	else {
		snprintf(sql_statement, BFSZ, "DELETE FROM %s WHERE %s=:uname", ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_FIELD_USER));
	}
	dmsg("SQL(%s)",sql_statement);
	if ( oparse(&cursor, sql_statement, -1, TRUE, 2) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Set :uname (declared in sql_statement above) to the one we want */
	if (!ini_value(ini,I_FIELD_DOMAIN)[0] && stricmp(ini_value(ini,I_APPEND_DOMAIN),"true")==0) snprintf(bf, BFSZ, "%s%s", uname, domain);
	else snprintf(bf, BFSZ, "%s", uname);

	if ( obndra(&cursor, ":uname", -1, bf, -1, SQLT_STR, -1, NULL, NULL, NULL, 0, NULL, NULL, 0, 0) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}
	if (ini_value(ini,I_FIELD_DOMAIN)[0]) {
		if ( obndra(&cursor, ":dom", -1, &domain[1], -1, SQLT_STR, -1, NULL, NULL, NULL, 0, NULL, NULL, 0, 0) ) {
			response("-ERR %s%s failed to bind :dom\n", uname, domain);
			return FALSE;
		}
	}

	/* Execute sql_statement */
	if ( oexec(&cursor) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Close cursor */
	if ( oclose(&cursor) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Commit changes to db */
	if ( ocom(&lda) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Output */
	response("+OK %s%s Deleted\n", uname, domain);
	return TRUE;
}

int oracle_search(char *string, int from, int max) {
	/*char search[BFSZ], domain[BFSZ];*/
	char temp[BFSZ], name[BFSZ], sql_statement[BFSZ];
	int field_count, user_ok, pass_ok, domain_ok, i, matches, iswild;
	pUInfo fields, ptr, uname, domain;
	
	/* Strip name + domain to name and domain */
	/*
	oa_stripname(string, search, domain, BFSZ);
	if ( ini_value(ini, I_DOMAIN)[0] ) {
		if ( !domain[0] ) {
			domain[0] = '@';
			ncpy(&domain[1], ini_value(ini, I_DOMAIN), BFSZ-1);
		}
	}
	*/
	/*else domain[0] = '\0';*/

	/* Ensure oracle connection */
	if ( !oracle_connect() ) {
		response("-ERR %s", oracle_error(NULL));
		return FALSE;
	}

	/* Find out names of all fields in oracle db, check for username and password */
	user_ok = pass_ok = FALSE;
	fields = oracle_fields(&field_count, &user_ok, &pass_ok, &domain_ok);
	if ( !fields ) return FALSE;
	if ( !user_ok ) {
		response("-ERR The config setting 'field_user' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_USER), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	if ( !pass_ok ) {
		response("-ERR The config setting 'field_pass' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_PASS), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}
	if ( !domain_ok ) {
		response("-ERR The config setting 'field_domain' is incorrect '%s' is not a field in table '%s' of db '%s'\n", ini_value(ini, I_FIELD_DOMAIN), ini_value(ini, I_ORACLE_TABLE), ini_value(ini, I_ORACLE_DB));
		return FALSE;
	}

	/* Open a cursor into db, this is used for all commands to db */
	if ( oopen(&cursor, &lda, NULL, -1, -1, NULL, -1) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Create comma sep list of all fields for sql_statement */
	temp[0] = '\0';
	for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next ) {
		if ( temp[0] != '\0' ) ncat(temp, ",", BFSZ);
		ncat(temp, ptr->field, BFSZ);
	}

	/* Build sql_statement */
	snprintf(sql_statement, BFSZ, "SELECT %s FROM %s", temp, ini_value(ini, I_ORACLE_TABLE));
	if ( oparse(&cursor, sql_statement, -1, TRUE, 2) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Execute sql_statement */
	if ( oexec(&cursor) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Setup buffers to retrieve results to */
	for ( i = 0, ptr = fields; i < field_count; i++, ptr = ptr->next ) {
		if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_USER)) ) uname = ptr;
		if ( !stricmp(ptr->field, ini_value(ini, I_FIELD_DOMAIN)) ) domain = ptr;
		odefin(&cursor, i+1, ptr->value, sizeof(ptr->value), SQLT_STR, -1, (sb2 *) -1, NULL, 0, 0, NULL, NULL);
	}

	/* Check if doing wild compare or strstr to choose matches */
	if ( strchr(string, '*') || strchr(string, '?') ) iswild = TRUE;
	else iswild = FALSE;
	
	/* Fetch results, count total users and ones matching search parameters */
	i = matches = 0;
	while ( TRUE ) {
		if ( ofetch(&cursor) ) {
			/* 1403 == No more data */
			if ( cursor.rc == 1403 ) break;
			/* 1405 == NULL occurred, 1406 == value truncated */
			else if ( cursor.rc != 1405 && cursor.rc != 1406 ) {
				response("-ERR %s", oracle_error(&cursor));
				return FALSE;
			}
		}
		if ( i >= from && (max == -1 || i < (max+from)) ) {
			if (ini_value(ini,I_FIELD_DOMAIN)[0]) snprintf(name, BFSZ, "%s@%s", uname->value, domain->value);
			else snprintf(name, BFSZ, "%s", uname->value);

			if ( iswild ) {
				if ( wild_match(name, string) ) {
					oracle_getinfo(fields,field_count,temp,BFSZ);
					response("+DATA %s %s\n", name, temp);
					matches++;
				}
			}
			else {
				if ( strstr(name, string) ) {
					oracle_getinfo(fields,field_count,temp,BFSZ);
					response("+DATA %s %s\n", name, temp);
					matches++;
				}
			}
		}
		i++;
	}

	/* Close cursor */
	if ( oclose(&cursor) ) {
		response("-ERR %s", oracle_error(&cursor));
		return FALSE;
	}

	/* Output */
	response("+OK Search Complete %d items found out of %d\n", matches, i);
	return TRUE;
}
